/*
    CharityWorld: ModelSwarm.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "ModelSwarm.h"
#import "Agent.h"
#import "Bank.h"
#import "Verbosity.h"
#import "DoubleSimple.h"
#ifndef __CYGWIN__
#  import "DoubleWarn.h"
#endif
#import "Shuffler.h"
#import <errno.h>
#import <random.h>
#import <objc/objc-api.h>
#import <stdlib.h>
#import <ctype.h>

static BOOL check_valid_int(char *str) {
  char *c;

  for(c = str; (*c) != '\0'; c++) {
    if(!isdigit((int)(*c))) {
      return NO;
    }
  }
  return YES;
}

static BOOL parse_coord_param(char **param, const char *paramName, int *x,
			      int *y, int xmax, int ymax, const char *alt) {
  char *xcoord, *ycoord, *paramcpy = NULL;

  if((*param) == NULL || strcmp((*param), "") == 0) {
    (*param) = "random";
  }
  if(alt != NULL && strcmp((*param), alt) == 0) {
    return YES;
  }
  
  if(strcmp((*param), "random") == 0) {
    xcoord = (*param);
    ycoord = (*param);
  }
  else {
    char *openbr, *closebr, *comma;

    paramcpy = strdup(*param);
    if(paramcpy == NULL) {
      perror("String duplication");
      abort();
    }
    openbr = paramcpy;
    closebr = paramcpy + strlen(paramcpy) - 1;
    comma = strchr(paramcpy, (int)',');
    if(((*openbr) != '[' || (*closebr) != ']' || comma == NULL)
       && ((*openbr) != '(' || (*closebr) != ')' || comma == NULL)) {
				// Round brackets added optionally
				// because on Swarm 2.1.1 supplied for
				// a PC, square brackets upset
				// Tcl/Tk. To avoid upsetting
				// LispArchiver, however, the whole
				// co-ordinate should be enclosed in
				// double quotes.
      fprintf(stderr, "Parse error in %s parameter: %s -- should be "
	      "a co-ordinate format [x,y] or the word \"random\"", 
	      paramName, (*param));
      if(alt != NULL) {
	fprintf(stderr, " or the word \"%s\"", alt);
      }
      fprintf(stderr, "\n");
      abort();
    }
    (*comma) = '\0';
    (*closebr) = '\0';
    xcoord = openbr + 1;
    ycoord = comma + 1;
    if(strcmp(xcoord, "random") != 0 && !check_valid_int(xcoord)) {
      fprintf(stderr, "Parse error in %s parameter: %s -- x"
	      " co-ordinate should be a number or the word \"random\"\n",
	      paramName, (*param));
      abort();
    }
    if(strcmp(ycoord, "random") != 0 && !check_valid_int(ycoord)) {
      fprintf(stderr, "Parse error in %s parameter: %s -- y"
	      " co-ordinate should be a number or the word \"random\"\n",
	      paramName, (*param));
      abort();
    }
  }
  if(strcmp(xcoord, "random") == 0) {
    (*x) = [uniformIntRand getIntegerWithMin: 0 withMax: xmax - 1];
  }
  else {
    (*x) = atoi(xcoord) - 1;
    if((*x) < 0 || (*x) >= xmax) {
      fprintf(stderr, "Error in %s parameter: %s -- x co-ordinate %d is out of"
	      " range [1, %d]\n", paramName, (*param), (*x) + 1, xmax);
      abort();
    }
  }
  if(strcmp(ycoord, "random") == 0) {
    (*y) = [uniformIntRand getIntegerWithMin: 0 withMax: ymax - 1];
  }
  else {
    (*y) = atoi(ycoord) - 1;
    if((*y) < 0 || (*y) >= ymax) {
      fprintf(stderr, "Error in %s parameter: %s -- y co-ordinate %d is out of"
	      " range [1, %d]\n", paramName, (*param), (*y) + 1, ymax);
      abort();
    }
  }
  if(paramcpy != NULL) free(paramcpy);
  return NO;
}


@implementation ModelSwarm

-(void)setGenerosityMinStr: (char *)minStr maxStr: (char *)maxStr {
  generosityMinStr = minStr;
  generosityMaxStr = maxStr;
}

-(void)setBegMinStr: (char *)minStr maxStr: (char *)maxStr {
  begMinStr = minStr;
  begMaxStr = maxStr;
}

-(void)setJackpotStr: (char *)jpStr {
  jackpotStr = jpStr;
}

-(void)setInitialWealthStr: (char *)iwStr {
  initialWealthStr = iwStr;
}

-(void)setAgentClassStr: (char *)acStr {
  agentClassStr = acStr;
}

-(void)setWinnerCoord: (char *)winnerc {
  winnerCoord = winnerc;
}

-(void)setLoserChooser: (char *)loserc {
  loserChooser = loserc;
}

-(void)setQueueClassStr: (char *)qcStr {
  queueClassStr = qcStr;
}

-(void)setEnvShape: (char *)shape neighbourhood: (char *)nbr xSize: (int)x
	     ySize: (int)y radius: (int)r {
  envShape = shape;
  nbrhood = nbr;
  xSize = x;
  ySize = y;
  radius = r;
}

-buildObjects {
  int x = 0, y = 0;
  int winnerX, winnerY, loserX, loserY;
  Agent *winningAgent, *dummyAgent, *loserAgent;
  int i;
  Class agentClass;
  Class queueClass;

  if([Verbosity showProgress]) printf("Building model objects");
  [super buildObjects];

  // Create the queues of beggars and donors using the class selected by
  // the user in the queueClassStr parameter

  queueClass = objc_get_class(queueClassStr);
  if(![queueClass conformsTo: @protocol(Queue)]) {
    raiseEvent(ProtocolViolation, "Queue class %s does not conform to Queue"
	       " protocol -- amend parameter file\n", queueClassStr);
  } 
  beggars = [queueClass create: self];
  donors = [queueClass create: self];

  generosityMax = [DoubleSimple parseKeepSameClass: generosityMaxStr];
  generosityMin = [DoubleSimple parseKeepSameClass: generosityMinStr];
  begMax = [DoubleSimple parseKeepSameClass: begMaxStr];
  begMin = [DoubleSimple parseKeepSameClass: begMinStr];
  jackpot = [DoubleSimple parseKeepSameClass: jackpotStr];
  initialWealth = [DoubleSimple parseKeepSameClass: initialWealthStr];

#ifndef __CYGWIN__
  if([jackpot isKindOf: [DoubleWarn class]] ||
     [begMin isKindOf: [DoubleWarn class]] ||
     [begMax isKindOf: [DoubleWarn class]] ||
     [generosityMin isKindOf: [DoubleWarn class]] ||
     [generosityMax isKindOf: [DoubleWarn class]]) {
    if([Verbosity showFloatingPointTraps]) {
      [DoubleWarn writeTraps: stdout];
    }
  }
#endif

  // Create other objects

  agents = [Array create: self setCount: xSize * ySize];
				// Array to store the agents in

  world = [Discrete2d createBegin: self];
  [world setSizeX: xSize Y: ySize];
  world = [world createEnd];
				// A grid to put the agents in a space
  if([Verbosity showProgress]) printf("Created world\n");
  
  // Create the agents, but first check that an appropriate class has been
  // selected for the agents in the agentClassStr parameter

  agentClass = objc_get_class(agentClassStr);
  if(![agentClass conformsTo: @protocol(Create)]) {
    raiseEvent(ProtocolViolation, "Agent class %s is not creatable -- amend "
	       "parameter file\n", agentClassStr);
  }
  dummyAgent = [agentClass create: scratchZone];
  if(![dummyAgent isKindOf: [Agent class]]) {
    raiseEvent(ProtocolViolation, "Agent class %s is not a kind of Agent "
	       "-- amend parameter file\n", agentClassStr);
  }
  [dummyAgent drop];

  for(x = 0, i = 0; x < xSize; x++) {
    for(y = 0; y < ySize; y++, i++) {
      Agent *a = [agentClass create: self model: self X: x Y: y];
				// Create the agent -- each agent needs to
				// know the model object so that it can tell
				// it to queue itself
      DoubleSimple *rand = [DoubleSimple new];

      if([generosityMin eq: generosityMax]) {
        [a setGenerosity: generosityMin];
      }
      else {
	[rand cset: [uniformDblRand getDoubleWithMin: [generosityMin get]
				    withMax: [generosityMax get]]];
        [a setGenerosity: rand];
      }
				// Set the generosity of the agent
      if([begMin eq: begMax]) {
        [a setBeg: begMin];
      }
      else {
	[rand cset: [uniformDblRand getDoubleWithMin: [begMin get]
				    withMax: [begMax get]]];
        [a setBeg: rand];
      }
				// Set how much the agent will beg
      [a addWealth: initialWealth];
				// Set the agent's initial wealth
      [agents atOffset: i put: a];
      [world putObject: a atX: x Y: y];
      [rand free];
    }
  }
  if([Verbosity showProgress]) printf("Created agents\n");

  // Create a neighbourhood for each agent. By default, this is a
  // Moore neighbourhood with the specified radius and the environment
  // topology is bounded.

  if(nbrhood == NULL || strcmp(nbrhood, "") == 0) nbrhood = "moore";
  if(envShape == NULL || strcmp(envShape, "") == 0) envShape = "planar";
  if(strcmp(nbrhood, "moore") != 0 && strcmp(nbrhood, "von-neumann") != 0) {
    fprintf(stderr, "Parse error in nbrhood parameter %s: should be either "
	    "\"moore\" or \"von-neumann\"\n", nbrhood);
    abort();
  }
  if(strcmp(envShape, "planar") != 0 && strcmp(envShape, "toroidal") != 0) {
    fprintf(stderr, "Parse error in envShape parameter %s: should be either "
	    "\"planar\" or \"toroidal\"\n", envShape);
    abort();
  }
  for(x = 0; x < xSize; x++) {
    for(y = 0; y < ySize; y++) {
      int xx, yy, xxx, yyy;
      Agent *a = [world getObjectAtX: x Y: y];
      for(xx = x - radius; xx <= x + radius; xx++) {
	if(strcmp(envShape, "planar") == 0
	   && (xx < 0 || xx >= xSize)) {
	  continue;
	}
	xxx = xx;
	while(xxx < 0) {
	  xxx += xSize;
	}
	xxx = xxx % xSize;
        for(yy = y - radius; yy <= y + radius; yy++) {
	  if(strcmp(envShape, "planar") == 0
	     && (yy < 0 || yy >= ySize)) {
	    continue;
	  }
	  yyy = yy;
	  while(yyy < 0) {
	    yyy += ySize;
	  }
	  yyy = yyy % ySize;
          if(xx == x && yy == y) continue;
	  if(strcmp(nbrhood, "von-neumann") == 0 && !(xx == x || yy == y)) {
	    continue;
	  }
	  if([Verbosity showConnections]) {
	    printf("Connecting agent (%d, %d) to (%d, %d)\n", x, y, xxx, yyy);
	  }
          [a addNeighbour: (Agent *)[world getObjectAtX: xxx Y: yyy]];
        }
      }
    }
  }
  if([Verbosity showProgress]) printf("Connected agents up\n");

  // Now we pick a winner for the jackpot, and tell every other agent to pay
  // the winner the ticket price (which is equal to the jackpot divided by
  // the number of agents). The winner will also have to pay the ticket price,
  // of course. Well, this happens if loserChooser == "all". If loserChooser
  // is a co-ordinate, then this loser pays the ticket price of all the
  // agents. Also, if loserChooser is "none" then the winner just gets the
  // jackpot.

  parse_coord_param(&winnerCoord, "winnerCoord", &winnerX, &winnerY, xSize,
		    ySize, NULL);

  winningAgent = [world getObjectAtX: winnerX Y: winnerY];

  if(loserChooser == NULL || strcmp(loserChooser, "") == 0) {
    loserChooser = "all";
  }

  if(strcmp(loserChooser, "none") == 0) {
    loserAgent = nil;
  }
  else if(parse_coord_param(&loserChooser, "loserChooser", &loserX, &loserY,
		       xSize, ySize, "all")) {
    loserAgent = nil;
  }
  else {
    while(loserX == winnerX && loserY == winnerY && (xSize > 1 || ySize > 1)) {
      loserX = [uniformIntRand getIntegerWithMin: 0 withMax: xSize - 1];
      loserY = [uniformIntRand getIntegerWithMin: 0 withMax: ySize - 1];
    }
    loserAgent = [world getObjectAtX: loserX Y: loserY];
  }    
    
  [Shuffler shuffleArray: agents];
  minWealth = maxWealth = [initialWealth get];
  for(i = 0; i < [agents getCount]; i++) {
    Agent *a = [agents atOffset: i];
    DoubleSimple *ticketPrice = [DoubleSimple new];
    
    [ticketPrice set: jackpot];
    [ticketPrice idiv: xSize * ySize];
    if(loserAgent == nil) {
      if(strcmp(loserChooser, "none") == 0) {
	if(a == winningAgent) {
	  [a addWealth: jackpot];
	  maxWealth += [jackpot get];
	}
      }
      else {
	[Bank transferNoTrace: ticketPrice from: a to: winningAgent];
	minWealth = -[ticketPrice get];
	maxWealth += [ticketPrice get];
      }
    }
    else if(a != winningAgent) {
      [Bank transferNoTrace: ticketPrice from: loserAgent to: winningAgent];
      minWealth -= [ticketPrice get];
      maxWealth += [ticketPrice get];
    }

    [ticketPrice free];
  }
  if([Verbosity showProgress]) printf("Initialised wealth distribution\n");

  // Ask all the agents if they want to reschedule after the wealth has been
  // redistributed.

  for(i = 0; i < [agents getCount]; i++) {
    Agent *a = [agents atOffset: i];
    [a rescheduleCheck];
  }
  if([Verbosity showDonorContents]) {
    printf("Content of donors list:\n[");
    [[donors getList] forEach: M(printLocation)];
    printf("]\n");
  }
  
  if([Verbosity showBeggarContents]) {
    printf("Content of beggars list:\n[");
    [[beggars getList] forEach: M(printLocation)];
    printf("]\n");
  }

  return self;
}

/*

buildActions

Build the schedule for the model. The schedule is pseudo-dynamic, and involves
calling the next method of this object each cycle.

*/

-buildActions {
  if([Verbosity showProgress]) printf("Building model schedule\n");
  [super buildActions];
  schedule = [Schedule createBegin: self];
  [schedule setRepeatInterval: 1];
  schedule = [schedule createEnd];
  [schedule at: 0 createActionTo: self message: M(next)];
  return self;
}

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [schedule activateIn: self];
  return [self getSwarmActivity];
}

/*

As mentioned above, the schedule is pseudo-dynamic, and the following
methods implement that. The next method takes the first beggar and
first donor off the beggar and donor schedule lists, and tells them to
step. Each agent class has various conditions under which it will put
itself on the beggar or donor list, which can be triggered under
various circumstances (such as giving or receiving money). The order
in which the agents are processed depends on the queuing protocol the
user selects for the scheduling lists in the queueClassStr parameter.

The queueBeggar method is called by the agent when they want to beg,
and the queueDonor method when they want to donate. Obviously since
there is a delay between being queued for an action and that action
occurring, and since in the mean time it is possible for the agent's
status to change, it is possible for the same agent to be on each list
each time.

*/

-next {
  static unsigned long long counter = 0;
  static BOOL terminated = NO;

  if(terminated) {
    printf("Terminated!\n");
    return self;
  }
  counter++;
  if([Verbosity showCycles]) {
    printf("--------------------------------------[ Cycle %09llu\n", counter);
  }
  if([beggars getCount] > 0) {
    Agent *a;

    a = [beggars remove];
    if([Verbosity showModelCallingAgent]) {
      printf("%llu: Stepping beggar (%d, %d)\n", counter, [a getX] + 1,
	     [a getY] + 1);
    }
    [a step];
  }
  else {
    if([Verbosity showBeggarEmpty]) {
      printf("%llu: No beggars left!\n", counter);
    }
  }
  if([donors getCount] > 0) {
    Agent *a;

    a = [donors remove];
    if([Verbosity showModelCallingAgent]) {
      printf("%llu: Stepping donor (%d, %d)\n", counter, [a getX] + 1,
	     [a getY] + 1);
    }
    [a step];
  }
  else {
    if([Verbosity showDonorEmpty]) {
      printf("%llu: No donors left!\n", counter);
    }
  }

  if([Verbosity showDonorContents]) {
    printf("Content of donors list:\n[");
    [[donors getList] forEach: M(printLocation)];
    printf("]\n");
  }
  
  if([Verbosity showBeggarContents]) {
    printf("Content of beggars list:\n[");
    [[beggars getList] forEach: M(printLocation)];
    printf("]\n");
  }
    
  if([donors getCount] == 0 && [beggars getCount] == 0 && !terminated) {
    int i;
    BOOL converged = YES;
    BOOL certain = YES;
    printf("%llu: No beggars or donors left!\n", counter);
    printf("Terminated\n");
    terminated = YES;
    for(i = 0; i < [agents getCount]; i++) {
      DoubleSimple *awealth = [[agents atOffset: i] getWealth];
      int j;
      for(j = i + 1; j < [agents getCount]; j++) {
	DoubleSimple *nwealth = [[agents atOffset: j] getWealth];

	if([awealth ne: nwealth]) {
	  converged = NO;
	  if([awealth eq: nwealth]) {
	    certain = NO;
	  }
	}
	else if(![awealth eq: nwealth]) {
	  certain = NO;
	}
	[nwealth free];
      }
      [awealth free];
    }
    printf("%s -- %s\n", converged ? "Converged" : "Not converged",
	   certain ? "certain" : "uncertain");
  }
  return self;
}

-(void)queueBeggar: (Agent *)a {
  if([Verbosity showQueueChange]) printf("Queueing beggar (%d, %d)\n",
					 [a getX] + 1, [a getY] + 1);
  [beggars add: a];
}

-(void)queueDonor: (Agent *)a {
  if([Verbosity showQueueChange]) printf("Queueing donor (%d, %d)\n", 
					 [a getX] + 1, [a getY] + 1);
  [donors add: a];
}

-getWorld {
  return world;
}

-getAgents {
  return agents;
}

-getDonors {
  return donors;
}

-getBeggars {
  return beggars;
}

-(int)getXSize {
  return xSize;
}

-(int)getYSize {
  return ySize;
}

-(DoubleSimple *)getJackpot {
  return [jackpot copy];
}

-(double)getJackpotForDisplayOnly {
  return [jackpot get];
}

-(DoubleSimple *)getInitialWealth {
  return [initialWealth copy];
}

-(double)getInitialWealthForDisplayOnly {
  return [initialWealth get];
}

-(double)getMinWealthForDisplayOnly {
  return minWealth;
}

-(double)getMaxWealthForDisplayOnly {
  return maxWealth;
}

-(void)drop {
  [generosityMax free];
  [generosityMin free];
  [begMax free];
  [begMin free];
  [jackpot free];
  [initialWealth free];
  [super drop];
  return;
}

@end
