/*
    CharityWorld: main.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "ObserverSwarm.h"
#import "Verbosity.h"
#import <errno.h>
#import <simtools.h>
#import <simtoolsgui.h>
#import <stdlib.h>
#import <string.h>
#import <sys/types.h>
#import <sys/stat.h>
#import <random.h>

#define DEFAULT_VERBOSITY_FILE "charity.verby"

int main(int argc, const char **argv) {
  ObserverSwarm *obsSwarm;
  const char *params;
  const char **fakeargv;
  const char *verbosityFile = DEFAULT_VERBOSITY_FILE;
  int fakeargc;
  int i, j;
  int swarmArgStart, paramArg;
  int verbosity = 0;
  int terminationTime = 0;
  int ndisplaywealthrastersave = 1;
  const char *wealthrasterfilestem = "wealthRaster";
  BOOL termSpecified = NO;
  BOOL blackAndWhite = NO;
  BOOL fileoutput = NO;
  BOOL savewealthraster = NO;
  BOOL showParseArgs = NO;
  const char *outputfile = "defaultScheduleOutput";
  struct stat fstat;

  // Supply a brief usage message if no arguments have been given

  if(argc == 1) {
    const char *helpargv[2];

    printf("Usage: %s [+f <output file>] [+v <verbosity>] "
	   "[+V <verbosity file>]\n\t[+t <termination time>] [+s <number of "
	   "times to save wealth raster>]\n\t[+S <stem for wealth raster "
	   "filenames>] [+b]\n\t<CharityWorld parameter file> "
	   "[Swarm options...]\n", argv[0]);
    printf("  +f <output file>\tSend the output to the specified file\n");
    printf("  +v <verbosity>\tLevel of verbosity\n");
    printf("  +V <verbosity file>\tVerbosity configuration file."
	   " (%s by default)\n", verbosityFile);
    printf("  +w\t\t\tShow parsing of command line arguments ");
    printf("  +t <termination time>\tNumber of cycles after which "
	   "to stop the simulation\n");
    printf("  +s <number>\t\tNumber of times to save the wealth raster\n");
    printf("  +S <file>\t\tFile to save wealth raster to\n");
    printf("  +b\t\t\tBlack and white rasters and graphs\n");
    printf("\nCharityWorld parameter file is a compulsory argument to this "
	   "command, and should\npoint to a Swarm scheme (.scm) file\n\n");
    printf("Swarm usage:\n\n");

    helpargv[0] = argv[0];
    helpargv[1] = "--help";

    initSwarm(2, helpargv);
    return 1;
  }

  // Cope with cases when this command is called without a parameter file

  if(argv[1][0] == '-') {
    if(strcmp(argv[1], "--version") != 0 && strcmp(argv[1], "-V") != 0) {
      printf("WARNING: No parameter file specified. For a full usage message, "
	     "type %s\n", argv[0]);
      printf("(If you think you have specified a parameter file, then note "
	     "that parameter\nfiles beginning with a - are not valid.)\n");
    }
    initSwarm(argc, argv);
    return 1;
  }
    

  // We've not used the Swarm argument handler, so we have to handle our
  // own arguments. These use a + instead of a - to indicate they are options,
  // and currently just two options are available that determine verbosity
  // -- the level of messages to print to stdout about the progress of the
  // model. +v <arg> takes an integer argument specifying the level of
  // verbosity (higher numbers mean more messages) +V <arg> takes a filename
  // argument -- the file referred to containing a list of message types that
  // are generated by the system, and a verbosity level at or above which the
  // messages are to be printed. The + options must appear first after the
  // command.

  printf("%s -- Flags:\n", argv[0]);
  for(i = 1; i < argc && argv[i][0] == '+'; i++) {
				// Loop from the first argument to the program
				// until the end of the arguments or the end
				// of arguments beginning +
    printf("%d: %s", i, argv[i]);
    switch(argv[i][1]) {
    case 'f':
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
	raiseEvent(InvalidArgument, "+f flag takes an argument");
      }
      fileoutput = YES;
      outputfile = argv[i];
      break;
    case 'v':
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
        raiseEvent(InvalidArgument, "+v flag takes an argument");
      }
      verbosity = atoi(argv[i]);
      break;
    case 'w':
      showParseArgs = YES;
      break;
    case 'V':
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
	raiseEvent(InvalidArgument, "+V flag takes an argument");
      }
      verbosityFile = argv[i];
      break;
    case 't':
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
        raiseEvent(InvalidArgument, "+t flag takes an argument");
      }
      terminationTime = atoi(argv[i]);
      termSpecified = YES;
      break;
    case 'b':
      blackAndWhite = YES;
      break;
    case 's':
      savewealthraster = YES;
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
	raiseEvent(InvalidArgument, "+s flag takes an argument");
      }
      ndisplaywealthrastersave = atoi(argv[i]);
      break;
    case 'S':
      savewealthraster = YES;
      i++;
      printf(" %s", argv[i]);
      if(argc <= i) {
	raiseEvent(InvalidArgument, "+S flag takes an argument");
      }
      wealthrasterfilestem = argv[i];
      break;
    default:
      raiseEvent(InvalidArgument, "%s flag not recognised", argv[i]);
    }
    printf("\n");
  }
  paramArg = i;			// Remember the last argument we were about
				// to use for this-program-specific arguments
				// -- this is the parameter file argument
  swarmArgStart = paramArg + 1;
				// The next argument will be an argument
				// for swarm-specific arguments

  if(showParseArgs) {
				// Display a message about parsing the
				// arguments to the program if the verbosity
				// level has been set appropriately
    printf("Parameter file at argument %d, Swarm args start at %d\n",
	   paramArg, swarmArgStart);
  }

  // The last argument before any swarm flags should be the parameter file

  if(argc <= paramArg) {
    raiseEvent(InvalidArgument, "You must specify the parameter file "
               "(after any charityWorld specific flags and before "
	       "any Swarm specific flags)");
  }

  // Now we create a fake argument vector to send to the Swarm initialisation

  params = argv[paramArg];
  if(params[0] == '-') {
    fprintf(stderr, "Parameter files beginning with a - are not valid (%s)\n",
	    params);
    abort();
  }
  if(stat(params, &fstat) != 0) {
    fprintf(stderr, "There is a problem with the parameter file:\n");
    perror(params);
    abort();
  }

  fakeargv = malloc((argc - swarmArgStart + 2) * sizeof(char *));
  if(fakeargv == NULL) {
    raiseEvent(OutOfMemory, "Could not allocate memory: %s", strerror(errno));
  }
  fakeargc = 0;
  for(i = 0, j = 0; i < argc; i++) {
    if(showParseArgs) printf("Analysing argument %s\n", argv[i]);
    if(i > 0 && i < swarmArgStart) {
      if(showParseArgs) {
	printf("Argument %s is not for Swarm\n", argv[i]);
      }
      continue;
    }
    if(showParseArgs) {
      printf("Argument %s is for Swarm\n", argv[i]);
    }
    fakeargv[j] = argv[i];
    fakeargc++;
    j++;
  }
  
  /////////////////////////////////////////////////////////////////////////////

  // Initialise Swarm

  initSwarm(fakeargc, fakeargv);

  // No methods should be called before the call to initSwarm. This
  // upsets PC implementations of Swarm.

  /////////////////////////////////////////////////////////////////////////////

  if(strcmp(verbosityFile, DEFAULT_VERBOSITY_FILE) != 0
     || stat(verbosityFile, &fstat) == 0) {
    [Verbosity initialise: verbosityFile];
				// Initialise the verbosity class if a
				// verbosity file has been specified using +V
				// or the default verbosity file exists.
				// N.B. If this file does not exist, the
				// default is not to show anything regardless
				// of the verbosity level.
  }
  [Verbosity setLevel: verbosity];
				// Set the verbosity level 

  if([Verbosity showProgress]) printf("Swarm initialised\n");
  printf("Seed: %u\n", [randomGenerator getInitialSeed]);

  // Start the model running

  if(swarmGUIMode == YES) {
    if(verbosity > 0) printf("Building observer\n");
    obsSwarm = [ObserverSwarm createBegin: globalZone];
    [obsSwarm setParameterFile: params];
    obsSwarm = [obsSwarm createEnd];
    if(termSpecified) [obsSwarm setTermTime: terminationTime];
    if(savewealthraster) [obsSwarm setWealthPNGSave: ndisplaywealthrastersave
				   fileStem: wealthrasterfilestem];
    [obsSwarm setBlackAndWhiteRasters: blackAndWhite];
    if(fileoutput) [obsSwarm setScheduleGraphFileOutput: outputfile];
    [obsSwarm buildObjects];
    [obsSwarm buildActions];
    [obsSwarm activateIn: nil];
    if([Verbosity showProgress]) printf("Starting the simulation\n");
    [obsSwarm go];
  }
  else {
    raiseEvent(InvalidArgument, "You've got to run this one in GUI mode!");
  }
  free(fakeargv);
  return 0;
}
